<div class="rst-versions" role="note" aria-label="{% trans %}Versions{% endtrans %}">
  <span class="rst-current-version" data-toggle="rst-current-version">
    {% if config.repo_name == 'GitHub' %}
        <span>
          <a href="{{ config.repo_url }}" class="fa fa-github" style="color: #fcfcfc"> GitHub</a>
        </span>
    {% elif config.repo_name == 'Bitbucket' %}
        <span>
          <a href="{{ config.repo_url }}" class="icon icon-bitbucket" style="color: #fcfcfc"> BitBucket</a>
        </span>
    {% elif config.repo_name == 'GitLab' %}
        <span>
          <a href="{{ config.repo_url }}" class="icon icon-gitlab" style="color: #fcfcfc"> GitLab</a>
        </span>
    {% endif %}
    {% if page.previous_page %}
      <span><a href="{{ page.previous_page.url|url }}" style="color: #fcfcfc">&laquo; {% trans %}Previous{% endtrans %}</a></span>
    {% endif %}
    {% if page.next_page %}
      <span><a href="{{ page.next_page.url|url }}" style="color: #fcfcfc">{% trans %}Next{% endtrans %} &raquo;</a></span>
    {% endif %}
  </span>
</div>
