<div role="navigation" aria-label="breadcrumbs navigation">
  <ul class="wy-breadcrumbs">
    <li><a href="{{ nav.homepage.url|url }}" class="icon icon-home" alt="{% trans %}Docs{% endtrans %}"></a> &raquo;</li>
    {%- if page %}
      {%- for doc in page.ancestors[::-1] %}
        {%- if doc.link %}
          <li><a href="{{ doc.link|e }}">{{ doc.title }}</a> &raquo;</li>
        {%- else %}
          <li>{{ doc.title }} &raquo;</li>
        {%- endif %}
      {%- endfor %}
    {%- endif %}
    {%- if page %}<li>{{ page.title }}</li>{%- endif %}
    <li class="wy-breadcrumbs-aside">
      {%- block repo %}
      {%- if page and page.edit_url %}
        <a href="{{ page.edit_url }}"
        {%- if config.repo_name|lower == 'github' %}
          class="icon icon-github"
        {%- elif config.repo_name|lower == 'bitbucket' %}
          class="icon icon-bitbucket"
        {%- elif config.repo_name|lower == 'gitlab' %}
          class="icon icon-gitlab"
        {%- endif %}> {% trans repo_name=config.repo_name %}Edit on {{ repo_name }}{% endtrans %}</a>
      {%- endif %}
      {%- endblock %}
    </li>
  </ul>
  {%- if config.theme.prev_next_buttons_location|lower in ['top', 'both']
        and page and (page.next_page or page.previous_page) %}
    <div class="rst-breadcrumbs-buttons" role="navigation" aria-label="{% trans %}Breadcrumb Navigation{% endtrans %}">
      {%- if page.previous_page %}
        <a href="{{ page.previous_page.url|url }}" class="btn btn-neutral float-left" title="{{ page.previous_page.title }}"><span class="icon icon-circle-arrow-left" aria-hidden="true"></span> {% trans %}Previous{% endtrans %}</a>
      {%- endif %}
      {%- if page.next_page %}
        <a href="{{ page.next_page.url|url }}" class="btn btn-neutral float-right" title="{{ page.next_page.title }}">{% trans %}Next{% endtrans %} <span class="icon icon-circle-arrow-right" aria-hidden="true"></span></a>
      {%- endif %}
    </div>
  {%- endif %}
  <hr/>
</div>

